/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef DEBUGGER_20061101_H_
#define DEBUGGER_20061101_H_

#include "Types.h"
#include "API.h"
#include <QString>
#include <QStringList>

class AnalyzerInterface;
class ArchProcessorInterface;
class BinaryInfo;
class Configuration;
class DebugEventHandlerInterface;
class DebuggerCoreInterface;
class DebuggerPluginInterface;
class FunctionDB;
class FunctionInfo;
class MemoryRegions;
class SessionFileInterface;
class State;
class SymbolManager;

class QByteArray;
class QDialog;
class QFileInfo;
class QWidget;

struct ExpressionError;

#include "MemRegion.h"
#include "BinaryInfo.h"

namespace edb {
	namespace v1 {

		// some useful objects
		EDB_EXPORT extern DebuggerCoreInterface	*debuggerBase;
		EDB_EXPORT extern QWidget				*debuggerUI;

		// get the current state object
		EDB_EXPORT State &currentState();

		// the symbol mananger
		EDB_EXPORT SymbolManager &symbolManager();

		// the memory region manager
		EDB_EXPORT MemoryRegions &memoryRegions();

		// the current arch processor
		EDB_EXPORT ArchProcessorInterface &archProcessor();

		// breakpoint managment
		EDB_EXPORT void removeBreakpoint(edb::address_t address);
		EDB_EXPORT void createBreakpoint(edb::address_t address);
		EDB_EXPORT void toggleBreakpoint(edb::address_t address);

		// change what the various views show
		EDB_EXPORT bool dumpDataRange(edb::address_t address, edb::address_t end_address, bool newTab);
		EDB_EXPORT bool dumpDataRange(edb::address_t address, edb::address_t end_address);	
		EDB_EXPORT bool dumpData(edb::address_t address, bool newTab);
		EDB_EXPORT bool dumpData(edb::address_t address);
		EDB_EXPORT bool dumpStack(edb::address_t address, bool ScrollTo);
		EDB_EXPORT bool dumpStack(edb::address_t address);
		EDB_EXPORT bool jumpToAddress(edb::address_t address);

		// ask the user for a value suitable for a register via an input box
		EDB_EXPORT bool getValueFromUser(edb::reg_t &value, const QString &title = QT_TR_NOOP("Input Value"));

		// ask the user for a binary string via an input box
		EDB_EXPORT bool getBinaryStringFromUser(QByteArray &value, const QString &title, int maxLength = 0);

		// determine if the given address is the starting point of an string, if so, s will contain it 
		// (formatted with C-style escape chars, so foundLength will have the original length of the string in chars).
		EDB_EXPORT bool getASCIIStringAtAddress(edb::address_t address, QString &s, int minLength, int maxLength, int &foundLength);
		EDB_EXPORT bool getUTF16StringAtAddress(edb::address_t address, QString &s, int minLength, int maxLength, int &foundLength);

		// list of loaded librarys
		EDB_EXPORT QStringList loadedLibraries();
		
		// the region which represents the main module of the application
		EDB_EXPORT MemRegion primaryCodeRegion();
		EDB_EXPORT MemRegion primaryDataRegion();

		// configuration
		EDB_EXPORT QDialog *dialogOptions();
		EDB_EXPORT const Configuration &config();

		// a numeric version of the current version suitable for integer comparison
		EDB_EXPORT ulong edbVersion();

		// symbol resolution
		EDB_EXPORT QString findFunctionSymbol(edb::address_t address);
		EDB_EXPORT QString findFunctionSymbol(edb::address_t address, const QString &defaultValue);
		EDB_EXPORT QString findFunctionSymbol(edb::address_t address, const QString &defaultValue, int *offset);

		EDB_EXPORT QString getProcessCurrentWorkingDirectory();
		EDB_EXPORT QString getProcessExecutable();
		EDB_EXPORT QStringList getProcessArguments();

		EDB_EXPORT edb::address_t getDWORD(edb::address_t address, bool &ok, ExpressionError &err, void *p);
		EDB_EXPORT edb::address_t getVariable(const QString &s, bool &ok, ExpressionError &err, void *p);

		EDB_EXPORT DebugEventHandlerInterface *setDebugEventHandler(DebugEventHandlerInterface *p);
		EDB_EXPORT DebugEventHandlerInterface *debugEventHandler();

		EDB_EXPORT AnalyzerInterface *setAnalyzer(AnalyzerInterface *p);
		EDB_EXPORT AnalyzerInterface *analyzer();
		
		EDB_EXPORT SessionFileInterface *setSessionFileHandler(SessionFileInterface *p);
		EDB_EXPORT SessionFileInterface *sessionFileHandler();

		EDB_EXPORT bool getInstructionBytes(edb::address_t address, quint8 *buf, int &size);

		EDB_EXPORT BinaryInfo *getBinaryInfo(const MemRegion &region);
		EDB_EXPORT const FunctionInfo *getFunctionInfo(const QString &functionName);
	
		EDB_EXPORT edb::address_t locateMainFunction();
		
		EDB_EXPORT bool registerPlugin(const QString &filename, QObject *plugin);
		EDB_EXPORT const QMap<QString, QObject *> &pluginList();
		EDB_EXPORT DebuggerPluginInterface *findPluginByName(const QString &name);

		EDB_EXPORT void reloadSymbols();		
		EDB_EXPORT void repaintCPUView();
		
		EDB_EXPORT void popValue(State &state);
		EDB_EXPORT void pushValue(State &state, edb::reg_t value);
		
		EDB_EXPORT void registerBinaryInfo(BinaryInfo::createFuncPtr fptr);
		EDB_EXPORT FunctionDB *setFunctionDB(FunctionDB *p);
		
		EDB_EXPORT bool overwriteCheck(edb::address_t address, unsigned int size);
		EDB_EXPORT void modifyBytes(edb::address_t address, unsigned int size, QByteArray &bytes, quint8 fill);
		
		EDB_EXPORT QByteArray getFileMD5(const QString &s);
		EDB_EXPORT QString basename(const QString &s);
		EDB_EXPORT QString symlinkTarget(const QString &s);
		EDB_EXPORT ulong intVersion(const QString &s);
		EDB_EXPORT QStringList parseCommandLine(const QString &cmdline);
		EDB_EXPORT edb::address_t stringToAddress(const QString &s, bool &ok);
		EDB_EXPORT QString formatBytes(const QByteArray &x);
		EDB_EXPORT QString formatPointer(edb::address_t p);
		
		EDB_EXPORT edb::address_t cpuSelectedAddress();
	}
}	
#endif
