/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "Bookmarks.h"
#include "Debugger.h"
#include "Expression.h"
#include "Debugger.h"

#include <QDockWidget>
#include <QGridLayout>
#include <QInputDialog>
#include <QTableWidget>
#include <QMainWindow>
#include <QMenu>
#include <QMessageBox>
#include <QPushButton>
#include <QShortcut>
#include <QSignalMapper>
#include <QHeaderView>

#include <memory>

//------------------------------------------------------------------------------
// Name: Bookmarks(
// Desc: 
//------------------------------------------------------------------------------
Bookmarks::Bookmarks() : m_Menu(0), m_ButtonAdd(0), m_ButtonDel(0), m_ButtonClear(0), m_Table(0) {
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::add() {
	bool ok;
    const QString text = QInputDialog::getText(edb::v1::debuggerUI, tr("Bookmark Address"), tr("Address:"), QLineEdit::Normal, QString(), &ok);
	
	if (ok && !text.isEmpty()) {
		Expression<edb::address_t> expr(text, edb::v1::getVariable, 0, edb::v1::getDWORD, 0);
		ExpressionError err;

		const edb::address_t address = expr.evaluateExpression(ok, err);
		if(ok) {
			addAddress(address);
		} else {
			QMessageBox::information(0, tr("Error In Address Expression!"), err.what());
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::del() {
	
	std::auto_ptr<QTableWidgetItem> item(m_Table->takeItem(m_Table->currentRow(), 0));
	m_Table->removeRow(m_Table->currentRow());
	if(item.get()) {
		bool ok;
		const QString s = item->text();
		const edb::address_t address = edb::v1::stringToAddress(s, ok);
		m_Entries.remove(address);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::clear() {
	m_Table->clearContents();
	m_Table->setRowCount(0);
	m_Entries.clear();
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::cellDoubleClicked(int row, int) {
	bool ok;
	QTableWidgetItem *const item = m_Table->item(row, 0);
	if(item != 0) {
		const QString s = item->text();
		const edb::address_t addr = edb::v1::stringToAddress(s, ok);
		if(ok) {
			edb::v1::jumpToAddress(addr);
		}
	}
}

//------------------------------------------------------------------------------
// Name: doShortcut(int index)
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::doShortcut(int index) {

	QTableWidgetItem *const item = m_Table->item(index, 0);
	
	if(item != 0) {
		emit cellDoubleClicked(index, 0);
	}
}

//------------------------------------------------------------------------------
// Name: menu(QWidget *parent)
// Desc: 
//------------------------------------------------------------------------------
QMenu *Bookmarks::menu(QWidget *parent) {
	if(m_Menu == 0) {
	
		// if we are dealing with a main window (and we are...)
		// add the dock object
		QMainWindow *const mainWindow = qobject_cast<QMainWindow *>(parent);
		if(mainWindow != 0) {
			
			// boiler plate button stuff
			m_ButtonAdd		= new QPushButton(tr("Add"));
			m_ButtonDel		= new QPushButton(tr("Del"));
			m_ButtonClear	= new QPushButton(tr("Clear"));
			m_Table			= new QTableWidget;
			
			m_Table->horizontalHeader()->setStretchLastSection(true);
			m_Table->setEditTriggers(QAbstractItemView::NoEditTriggers);
			m_Table->setSelectionBehavior(QAbstractItemView::SelectRows);
			m_Table->verticalHeader()->hide();
			m_Table->setColumnCount(2);
			m_Table->setHorizontalHeaderItem(0, new QTableWidgetItem(tr("Address")));
			m_Table->setHorizontalHeaderItem(1, new QTableWidgetItem(tr("Comment")));
			m_Table->setContextMenuPolicy(Qt::CustomContextMenu);
			connect(m_Table, SIGNAL(customContextMenuRequested(const QPoint &)), SLOT(customContextMenuRequested(const QPoint &)));
			
			// events..
			connect(m_ButtonAdd, SIGNAL(clicked()), this, SLOT(add()));
			connect(m_ButtonDel, SIGNAL(clicked()), this, SLOT(del()));
			connect(m_ButtonClear, SIGNAL(clicked()), this, SLOT(clear()));
			connect(m_Table, SIGNAL(cellDoubleClicked(int,int)), this, SLOT(cellDoubleClicked(int,int)));

			// layouts
			QBoxLayout *const buttonLayout = new QBoxLayout(QBoxLayout::LeftToRight);
			buttonLayout->addWidget(m_ButtonAdd);
			buttonLayout->addWidget(m_ButtonDel);
			buttonLayout->addWidget(m_ButtonClear);

			QBoxLayout *const layout = new QBoxLayout(QBoxLayout::TopToBottom);
			layout->addWidget(m_Table);
			layout->addLayout(buttonLayout);

			QWidget *const widget = new QWidget(parent);
			widget->setLayout(layout);

			// make the dock widget and _name_ it, it is important to name it so
			// that it's state is saved in the GUI info
			QDockWidget *const dockWidget = new QDockWidget(tr("Bookmarks"), parent);
			dockWidget->setObjectName(QString::fromUtf8("Bookmarks"));
			dockWidget->setWidget(widget);

			// add it to the dock
			mainWindow->addDockWidget(Qt::RightDockWidgetArea, dockWidget);

			// make the menu and add the show/hide toggle for the widget
			m_Menu = new QMenu(tr("Bookmarks"), parent);
			m_Menu->addAction(dockWidget->toggleViewAction());
			
			m_SignalMapper = new QSignalMapper(this);
			
			for(int i = 0; i < 10; ++i) {
				// create an action and attach it to the signal mapper
				QShortcut *const action = new QShortcut(QKeySequence(tr("Ctrl+%n", "", i)), mainWindow);
								
				if(i == 0) {
					m_SignalMapper->setMapping(action, 9);
				} else {
					m_SignalMapper->setMapping(action, i - 1);
				}
				
				connect(action, SIGNAL(activated()), m_SignalMapper, SLOT(map()));
			}
			
			// connect the parametized signal to the slot..phew finally
			connect(m_SignalMapper, SIGNAL(mapped(int)), this, SLOT(doShortcut(int)));
		}
	}

	return m_Menu;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
QList<QAction *> Bookmarks::cpuViewContextMenu() {
	
	QList<QAction *> ret;
	
	QAction *const actFind = new QAction(tr("Add &Bookmark"), this);
	connect(actFind, SIGNAL(triggered()), this, SLOT(mnuBookmarkCPU()));
	ret << actFind;

	return ret;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::addAddress(edb::address_t address) {
	if(!m_Entries.contains(address)) {
		QTableWidgetItem *const new_item = new QTableWidgetItem(edb::v1::formatPointer(address));
		const int row = m_Table->rowCount();
		m_Table->setRowCount(row + 1);
		m_Table->setItem(row, 0, new_item);
		m_Table->resizeColumnToContents(0);
		m_Entries.insert(address);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::mnuBookmarkCPU() {
	addAddress(edb::v1::cpuSelectedAddress());
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void Bookmarks::customContextMenuRequested(const QPoint &pos) {

	QMenu menu;
	QAction *const actionAdd		= menu.addAction(tr("&Add Address"));
	QAction *const actionDel		= menu.addAction(tr("&Delete Address"));
	QAction *const actionClear		= menu.addAction(tr("&Clear"));
	menu.addSeparator();
	QAction *const actionComment	= menu.addAction(tr("&Set Comment"));
	QAction *const chosen = menu.exec(m_Table->mapToGlobal(pos));

	if(chosen == actionAdd) {
		add();
	} else if(chosen == actionDel) {
		del();
	} else if(chosen == actionClear) {
		clear();
	} else if(chosen == actionComment) {
		bool ok;
		
		QString current_comment;
		QTableWidgetItem *const item = m_Table->item(m_Table->currentRow(), 1);
		if(item != 0) {
			current_comment = item->text();
		}
		
	    const QString text = QInputDialog::getText(m_Table, tr("Comment"), tr("Set Comment:"), QLineEdit::Normal, current_comment, &ok);
		if(ok) {
			m_Table->setItem(m_Table->currentRow(), 1, new QTableWidgetItem(text));
		}
	}
}

Q_EXPORT_PLUGIN2(Bookmarks, Bookmarks)
