/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DialogHeap.h"
#include "Configuration.h"
#include "DebuggerCoreInterface.h"
#include "Debugger.h"
#include "MemoryRegions.h"
#include "SymbolManager.h"
#include "Util.h"

#include <QtDebug>
#include <QHeaderView>
#include <QMessageBox>
#include <QVector>
#include <QString>
#include <QFileInfo>
#include <QSortFilterProxyModel>
#include <QStack>
#include <QtDebug>

#include <boost/bind.hpp>

#include <algorithm>

#ifdef ENABLE_GRAPH
#include "GraphWidget.h"
#include <graphviz/gvc.h>
#endif

#ifdef USE_QT_CONCURRENT
#include <QtConcurrentMap>
#endif

#define PREV_INUSE 0x1
#define IS_MMAPPED 0x2
#define NON_MAIN_ARENA 0x4
#define SIZE_BITS (PREV_INUSE|IS_MMAPPED|NON_MAIN_ARENA)

#define next_chunk(p, c) ((p) + ((c).chunksize()))
#define prev_chunk(p, c) ((p) - ((c).prev_size)))


struct malloc_chunk {
	ulong	prev_size;	/* Size of previous chunk (if free).  */
	ulong	size;		/* Size in bytes, including overhead. */

	struct malloc_chunk* fd;	/* double links -- used only if free. */
	struct malloc_chunk* bk;

	ulong chunksize() const	{ return size & ~(SIZE_BITS); }
	bool prev_inuse() const	{ return size & PREV_INUSE; }
};


//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
DialogHeap::DialogHeap(QWidget *parent) : QDialog(parent) {
	ui.setupUi(this);
	
	m_Model = new ResultViewModel(this);
	ui.tableView->setModel(m_Model);
	
	ui.tableView->verticalHeader()->hide();
	ui.tableView->horizontalHeader()->setStretchLastSection(true);
	ui.tableView->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
	
#ifndef ENABLE_GRAPH
	ui.btnGraph->setEnabled(false);
#endif
}

//------------------------------------------------------------------------------
// Name: showEvent(QShowEvent *)
// Desc: 
//------------------------------------------------------------------------------
void DialogHeap::showEvent(QShowEvent *) {
	m_Model->clearResults();
	ui.progressBar->setValue(0);
}

//------------------------------------------------------------------------------
// Name: on_resultTable_cellDoubleClicked(int, int)
// Desc: 
//------------------------------------------------------------------------------
void DialogHeap::on_tableView_doubleClicked(const QModelIndex &index) {

	// NOTE: remember that if we use a sort filter, we need to map the indexes
	// to get at the data we need

	const Result *const item = static_cast<Result *>(index.internalPointer());
	
	if(item != 0) {
		edb::v1::dumpDataRange(item->block, item->block + item->size, false);
	}
}

//------------------------------------------------------------------------------
// Name: getLibraryNames(QString &libcName, QString &ldName) const
// Desc: 
//------------------------------------------------------------------------------
void DialogHeap::getLibraryNames(QString &libcName, QString &ldName) const {
	const QStringList libs = edb::v1::loadedLibraries();

	Q_FOREACH(const QString &s, libs) {
		if(!ldName.isEmpty() && !libcName.isEmpty()) {
			break;
		}

		const QFileInfo fileinfo(s);

		// this tries its best to cover all possible libc library versioning 
		// possibilities we need to find out if this is 100% accurate, so far 
		// seems correct based on my system
		
		if(fileinfo.completeBaseName().startsWith("libc-")) {
			libcName = fileinfo.completeBaseName() + "." + fileinfo.suffix();
			qDebug() << "[Heap Analyzer] libc library appears to be:" << libcName;
			continue;
		}
		
		if(fileinfo.completeBaseName().startsWith("libc.so")) {
			libcName = fileinfo.completeBaseName() + "." + fileinfo.suffix();
			qDebug() << "[Heap Analyzer] libc library appears to be:" << libcName;
			continue;
		}

		if(fileinfo.completeBaseName().startsWith("ld-")) {
			ldName = fileinfo.completeBaseName() + "." + fileinfo.suffix();
			qDebug() << "[Heap Analyzer] ld library appears to be:" << ldName;
			continue;
		}
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
void DialogHeap::processPotentialPointer(const QMap<edb::address_t, edb::address_t> &targets, Result &result) {
	if(result.data.isEmpty()) {
		edb::address_t pointer;
		edb::address_t block_start	= blockStart(result);
		edb::address_t block_end	= block_start + result.size;
		
		while(block_start < block_end) {
			
			if(edb::v1::debuggerBase->readBytes(block_start, &pointer, sizeof(pointer))) {
				QMap<edb::address_t, edb::address_t>::const_iterator it = targets.find(pointer);
				if(it != targets.end()) {
				#if QT_POINTER_SIZE == 4
					result.data += QString("dword ptr [%1] |").arg(edb::v1::formatPointer(it.key()));
				#elif QT_POINTER_SIZE == 8
					result.data += QString("qword ptr [%1] |").arg(edb::v1::formatPointer(it.key()));
				#endif
				result.points_to.push_back(it.value());
				}
			}
			block_start += sizeof(edb::address_t);
		}
		result.data.truncate(result.data.size() - 2);
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::address_t DialogHeap::blockStart(edb::address_t pointer) {
	return pointer + sizeof(struct malloc_chunk *) * 2;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::address_t DialogHeap::blockStart(const Result &result) {
	return blockStart(result.block);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::address_t DialogHeap::blockStart(const Result *result) {
	return blockStart(*result);
}

//------------------------------------------------------------------------------
// Name: detectPointers()
// Desc: 
//------------------------------------------------------------------------------
void DialogHeap::detectPointers() {
	
	qDebug() << "[Heap Analyzer] detecting pointers in heap blocks";
	
	QMap<edb::address_t, edb::address_t> targets;
	QVector<Result> &results = m_Model->results();
	
	// the potential targets
	qDebug() << "[Heap Analyzer] collecting possible targets addresses";
	Q_FOREACH(const Result &result, results) {
		edb::address_t block_start	= blockStart(result);
		edb::address_t block_end	= block_start + result.size;
		while(block_start < block_end) {
			targets.insert(block_start, result.block);
			block_start += sizeof(edb::address_t);
		}
	}
	
#ifdef USE_QT_CONCURRENT
	QtConcurrent::blockingMap(results, boost::bind(&DialogHeap::processPotentialPointer, this, targets, _1));
#else
	std::for_each(results.begin(), results.end(), boost::bind(&DialogHeap::processPotentialPointer, this, targets, _1));
#endif

	m_Model->update();
}

//------------------------------------------------------------------------------
// Name: collectBlocks(edb::address_t start_address, edb::address_t end_address)
// Desc: 
//------------------------------------------------------------------------------
void DialogHeap::collectBlocks(edb::address_t start_address, edb::address_t end_address) {
	m_Model->clearResults();
	
	const int minStringLength = edb::v1::config().minStringLength;
	
	if(start_address != edb::nullptr && end_address != edb::nullptr) {
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD) || defined(Q_OS_OPENBSD)
		malloc_chunk currentChunk;
		malloc_chunk nextChunk;
		edb::address_t currentChunkAddress = start_address;

		m_Model->setUpdatesEnabled(false);

		const edb::address_t how_many = end_address - start_address;
		while(currentChunkAddress != end_address) {
			// read in the current chunk..
			edb::v1::debuggerBase->readBytes(currentChunkAddress, &currentChunk, sizeof(currentChunk));

			// figure out the address of the next chunk
			const edb::address_t nextChunkAddress = next_chunk(currentChunkAddress, currentChunk);

			// is this the last chunk (if so, it's the 'top')
			if(nextChunkAddress == end_address) {

				const Result r(
					currentChunkAddress, 
					currentChunk.chunksize(), 
					tr("Top"));
				
				m_Model->addResult(r);

			} else {

				// make sure we aren't following a broken heap...
				if(nextChunkAddress > end_address || nextChunkAddress < start_address) {
					break;
				}
				
				QString data;

				// read in the next chunk
				edb::v1::debuggerBase->readBytes(nextChunkAddress, &nextChunk, sizeof(nextChunk));

				// if this block is a container for an ascii string, display it...
				// there is a lot of room for improvement here, but it's a start
				QString asciiData;
				QString utf16Data;
				int asciisz;
				int utf16sz;
				if(edb::v1::getASCIIStringAtAddress(
						blockStart(currentChunkAddress),
						asciiData, 
						minStringLength,
						currentChunk.chunksize(),
						asciisz)) {
						
					data = "ASCII \"" + asciiData + "\"";
				} else if(edb::v1::getUTF16StringAtAddress(
						blockStart(currentChunkAddress),
						utf16Data,
						minStringLength,
						currentChunk.chunksize(),
						utf16sz)) {
					data = "UTF-16 \"" + utf16Data + "\"";
				}

				const Result r(
					currentChunkAddress, 
					currentChunk.chunksize() + sizeof(unsigned int), 
					nextChunk.prev_inuse() ? tr("Busy") : tr("Free"),
					data);
				
				m_Model->addResult(r);	
			}

			// avoif self referencing blocks
			if(currentChunkAddress == nextChunkAddress) {
				break;
			}

			currentChunkAddress = nextChunkAddress;

			ui.progressBar->setValue(util::percentage(currentChunkAddress - start_address, how_many));
		}
		
		detectPointers();
		m_Model->setUpdatesEnabled(true);
		
		
#else
	#error "Unsupported Platform"
#endif
	}
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
edb::address_t DialogHeap::findHeapStartHeuristic(edb::address_t end_address, size_t offset) const {
	edb::address_t start_address = end_address - offset;
	
#if defined(EDB_X86_64)
	edb::address_t heap_symbol = start_address - (sizeof(edb::address_t) + sizeof(edb::address_t) + sizeof(edb::address_t) + sizeof(edb::address_t));
#else
	edb::address_t heap_symbol = start_address - (sizeof(edb::address_t) + sizeof(edb::address_t) + sizeof(edb::address_t) + sizeof(unsigned int));
#endif

	edb::address_t test_addr;		
	edb::v1::debuggerBase->readBytes(heap_symbol, &test_addr, sizeof(test_addr));

	if(test_addr != edb::v1::debuggerBase->pageSize()) {
		return 0;
	}
	
	return start_address;
}

//------------------------------------------------------------------------------
// Name: doFind()
// Desc: 
//------------------------------------------------------------------------------
void DialogHeap::doFind() {
	// get both the libc and ld symbols of __curbrk
	// this will be the 'before/after libc' addresses
	
	// we should get them from the regions we see
	const SymbolManager::Symbol *s;

	edb::address_t start_address	= edb::nullptr;
	edb::address_t end_address		= edb::nullptr;

	QString libcName;
	QString ldName;

	getLibraryNames(libcName, ldName);
#if defined(Q_OS_LINUX) || defined(Q_OS_FREEBSD) || defined(Q_OS_OPENBSD)
	s = edb::v1::symbolManager().find(libcName + "::__curbrk");
	if(s != 0) {
		end_address = s->address;
	} else {
		QMessageBox::information(this, tr("__curbrk symbol not found in libc"), tr("Could not find the symbol for <strong>__curbrk</strong> in your libc, perhaps you need to regenerate your symbols?"));
		qDebug() << "[Heap Analyzer] __curbrk symbol not found in libc";
		return;
	}

	s = edb::v1::symbolManager().find(ldName + "::__curbrk");
	if(s != 0) {
		start_address = s->address;
	} else {
	
		qDebug() << "[Heap Analyzer] __curbrk symbol not found in ld, falling back on heuristic! This may or may not work.";
		
		for(edb::address_t offset = 0x0000; offset != 0x1000; offset += sizeof(edb::address_t)) {
			start_address = findHeapStartHeuristic(end_address, offset);
			if(start_address != edb::nullptr) {
				break;
			}
		}
	
		// ok, I give up
		if(start_address == edb::nullptr) {
			QMessageBox::information(this, tr("Could not calculate heap start"), tr("Failed to calculate the begining of the heap."));
			return;
		}
	}

	qDebug("[Heap Analyzer] heap start symbol : " EDB_FMT_PTR, start_address);
	qDebug("[Heap Analyzer] heap end symbol   : " EDB_FMT_PTR, end_address);
	
	// read the contents of those symbols
	edb::v1::debuggerBase->readBytes(end_address, &end_address, sizeof(end_address));
	edb::v1::debuggerBase->readBytes(start_address, &start_address, sizeof(start_address));
#else
	#error "Unsupported Platform"
#endif

	qDebug("[Heap Analyzer] heap start : " EDB_FMT_PTR, start_address);
	qDebug("[Heap Analyzer] heap end   : " EDB_FMT_PTR, end_address);

	collectBlocks(start_address, end_address);
}

//------------------------------------------------------------------------------
// Name: on_btnFind_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogHeap::on_btnFind_clicked() {
	ui.btnFind->setEnabled(false);
	ui.progressBar->setValue(0);
	doFind();
	ui.progressBar->setValue(100);
	ui.btnFind->setEnabled(true);
}

//------------------------------------------------------------------------------
// Name: on_btnGraph_clicked()
// Desc: 
//------------------------------------------------------------------------------
void DialogHeap::on_btnGraph_clicked() {
#ifdef ENABLE_GRAPH
	GVC_t *gvc = gvContext();
	graph_t *g = agopen(const_cast<char*>("g"), AGDIGRAPH);
	
	const QVector<Result> &results = m_Model->results();
	
	QMap<edb::address_t, node_t *>			nodes;
	QMap<edb::address_t, const Result *>	result_map;
	QStack<const Result *>					result_stack;
	QSet<const Result *>					seen_results;


	// first we make a nice index for our results, this is likely redundant, 
	// but won't take long
	Q_FOREACH(const Result &result, results) {
		result_map.insert(result.block, &result);
	}
	
	// seed our search with the selected blocks
	const QItemSelectionModel *const selModel = ui.tableView->selectionModel();
	const QModelIndexList sel = selModel->selectedRows();
	if(sel.size() != 0) {
		Q_FOREACH(QModelIndex index, sel) {
			const Result *const item = static_cast<Result *>(index.internalPointer());
				result_stack.push(item);
				seen_results.insert(item);
		}
	}
	
	while(!result_stack.isEmpty()) {
		const Result *result = result_stack.pop();
		node_t *n = agnode(g, const_cast<char*>(qPrintable(edb::v1::formatPointer(result->block))));
		if(result->type == tr("Busy")) {
			agsafeset(n, const_cast<char*>("fillcolor"), const_cast<char*>("green"), const_cast<char*>(""));
		} else {
			agsafeset(n, const_cast<char*>("fillcolor"), const_cast<char*>("red"), const_cast<char*>(""));
		}
		agsafeset(n, const_cast<char*>("style"), const_cast<char*>("filled"), const_cast<char*>(""));
		//agsafeset(n, const_cast<char*>("shape"), const_cast<char*>("box"), const_cast<char*>(""));
				
		nodes.insert(result->block, n);

		Q_FOREACH(edb::address_t pointer, result->points_to) {
			const Result *next_result = result_map[pointer];
			if(!seen_results.contains(next_result)) {
				seen_results.insert(next_result);
				result_stack.push(next_result);
			}
		}
	}
	
	Q_FOREACH(const Result *result, result_map) {
		const edb::address_t addr = result->block;
		if(nodes.contains(addr)) {
			Q_FOREACH(edb::address_t pointer, result->points_to) {
				agedge(g, nodes[addr], nodes[pointer]);
			}
		}
	}

	GraphWidget *graph = new GraphWidget(gvc, g, "dot");
	graph->show();
	
	agclose(g);
	gvFreeContext(gvc);
#endif
}
